// area.c -- routines to assist with inserting "areas" into FG terrain
//
// Written by Curtis Olson, started March 1998.
//
// Copyright (C) 1998  Curtis L. Olson  - curt@me.umn.edu
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: area.cxx,v 1.5 1999/10/09 12:52:17 curt Exp $
//


#include <math.h>
#include <stdio.h>

#include <Include/fg_constants.h>

#include <Build/construct_types.hxx>

#include "area.hxx"
#include "point2d.hxx"


// calc new x, y for a rotation
double rot_x(double x, double y, double theta) {
    return ( x * cos(theta) + y * sin(theta) );
}


// calc new x, y for a rotation
double rot_y(double x, double y, double theta) {
    return ( -x * sin(theta) + y * cos(theta) );
}


// calc new lon/lat given starting lon/lat, and offset radial, and
// distance.  NOTE: distance is specified in meters (and converted
// internally to radians)
Point3D calc_lon_lat( const Point3D& orig, Point3D offset ) {
    Point3D result;

    // printf("calc_lon_lat()  offset.theta = %.2f offset.dist = %.2f\n",
    //        offset.theta, offset.dist);

    offset.setx( offset.x() * METER_TO_NM * NM_TO_RAD );

    result.sety( asin( sin(orig.y()) * cos(offset.x()) + 
		       cos(orig.y()) * sin(offset.x()) * cos(offset.y()) ) );

    if ( cos(result.y()) < FG_EPSILON ) {
        result.setx( orig.x() );      // endpoint a pole
    } else {
        result.setx( 
	    fmod(orig.x() - asin( sin(offset.y()) * sin(offset.x()) / 
				  cos(result.y()) ) + FG_PI, FG_2PI) - FG_PI );
    }

    return result;
}


FGPolygon batch_cart_to_polar_2d( const FGPolygon& in_list ) {
    FGPolygon out_list;
    Point3D p;

    out_list.erase();
    for ( int i = 0; i < (int)in_list.contour_size( 0 ); ++i ) {
	p = cart_to_polar_2d( in_list.get_pt( 0, i ) );
	out_list.add_node( 0, p );
    }

    return out_list;
}


// given a set of 2d coordinates relative to a center point, and the
// lon, lat of that center point (specified in degrees), as well as a
// potential orientation angle, generate the corresponding lon and lat
// of the original 2d verticies.
FGPolygon gen_area(Point3D origin, double angle, const FGPolygon& cart_list) {
    FGPolygon rad_list;
    FGPolygon result_list;
    Point3D p;

    Point3D origin_rad( origin.x() * DEG_TO_RAD, origin.y() * DEG_TO_RAD, 0 );
	
    // convert to polar coordinates
    rad_list = batch_cart_to_polar_2d(cart_list);

    // display points
    // cout << "converted to polar" << endl;
    // for ( int i = 0; i < rad_list.contour_size( 0 ); ++i ) {
    //     cout << "  " << rad_list.get_pt(0, i) << endl;
    // }

    // rotate by specified angle
    // cout << "Rotating points by " << angle << endl;
    for ( int i = 0; i < rad_list.contour_size( 0 ); ++i) {
	p = rad_list.get_pt( 0, i );
	double theta = p.y() - angle;
        while ( theta < FG_2PI ) {
            theta += FG_2PI;
	}
	p.sety( theta );
	rad_list.set_pt( 0, i, p );
	// cout << "  " << p << endl;
    }

    // find actual lon,lat of coordinates
    // cout << "convert to lon, lat relative to " << origin << endl;
    for ( int i = 0; i < (int)rad_list.contour_size( 0 ); ++i ) {
	p = calc_lon_lat(origin_rad, rad_list.get_pt(0, i) );
	// convert from radians to degress
	p.setx( p.x() * RAD_TO_DEG );
	p.sety( p.y() * RAD_TO_DEG );
	// cout << "  " << p << endl;
	result_list.add_node( 0, p );
    }

    return result_list;
}


// generate an area for a runway
FGPolygon gen_runway_area( double lon, double lat, double heading, 
			   double length, double width) 
{
    FGPolygon result_list;
    FGPolygon tmp_list;

    double l, w;

    /*
    printf("runway: lon = %.2f lat = %.2f hdg = %.2f len = %.2f width = %.2f\n",
	   lon, lat, heading, length, width);
    */

    Point3D origin(lon, lat, 0);
    l = length / 2.0;
    w = width / 2.0;

    // generate untransformed runway area vertices
    tmp_list.add_node( 0, Point3D( l, w, 0 ) );
    tmp_list.add_node( 0, Point3D( l, -w, 0 ) );
    tmp_list.add_node( 0, Point3D( -l, -w, 0 ) );
    tmp_list.add_node( 0, Point3D( -l, w, 0 ) );

    // display points
    // cout << "Untransformed, unrotated runway" << endl;
    // for ( int i = 0; i < tmp_list.contour_size( 0 ); ++i ) {
    //    cout << "  " << tmp_list.get_pt(0, i) << endl;
    // }

    // rotate, transform, and convert points to lon, lat in degrees
    result_list = gen_area(origin, heading, tmp_list);

    // display points
    // cout << "Results in radians." << endl;
    // for ( int i = 0; i < result_list.contour_size( 0 ); ++i ) {
    //     cout << "  " << result_list.get_pt(0, i) << endl;
    // }

    return result_list;
}


// generate an area for a runway and include midpoints
FGPolygon gen_runway_w_mid( double lon, double lat, double heading, 
			   double length, double width) 
{
    FGPolygon result_list;
    FGPolygon tmp_list;

    double l, w;

    /*
    printf("runway: lon = %.2f lat = %.2f hdg = %.2f len = %.2f width = %.2f\n",
	   lon, lat, heading, length, width);
    */

    Point3D origin(lon, lat, 0);
    l = length / 2.0;
    w = width / 2.0;

    // generate untransformed runway area vertices
    tmp_list.add_node( 0, Point3D( l, w, 0 ) );
    tmp_list.add_node( 0, Point3D( l, -w, 0 ) );
    tmp_list.add_node( 0, Point3D( 0, -w, 0 ) );
    tmp_list.add_node( 0, Point3D( -l, -w, 0 ) );
    tmp_list.add_node( 0, Point3D( -l, w, 0 ) );
    tmp_list.add_node( 0, Point3D( 0, w, 0 ) );

    // display points
    // cout << "Untransformed, unrotated runway" << endl;
    // for ( int i = 0; i < tmp_list.contour_size( 0 ); ++i ) {
    //    cout << "  " << tmp_list.get_pt(0, i) << endl;
    // }

    // rotate, transform, and convert points to lon, lat in degrees
    result_list = gen_area(origin, heading, tmp_list);

    // display points
    // cout << "Results in radians." << endl;
    // for ( int i = 0; i < result_list.contour_size( 0 ); ++i ) {
    //     cout << "  " << result_list.get_pt(0, i) << endl;
    // }

    return result_list;
}
