// build.hxx -- routines to build polygon model of an airport from the runway
//              definition
//
// Written by Curtis Olson, started September 1999.
//
// Copyright (C) 1999  Curtis L. Olson  - curt@flightgear.org
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// $Id: build.cxx,v 1.7 1999/10/19 18:34:58 curt Exp $
//


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <Include/compiler.h>

#include <stdio.h>
#include <stdlib.h>		// for atoi() atof()
#include <time.h>

#include <list>
#include STL_STRING

#include <sg.h>			// plib include

#include <Bucket/newbucket.hxx>
#include <Include/fg_constants.h>
#include <Math/fg_geodesy.hxx>

#include <Array/array.hxx>
#include <Build/texcoord.hxx>
#include <Build/trinodes.hxx>
#include <Polygon/polygon.hxx>

#include "area.hxx"
#include "convex_hull.hxx"
#include "point2d.hxx"
#include "scenery_version.hxx"


typedef vector < int_list > group_list;
typedef group_list::iterator group_list_iterator;
typedef group_list::const_iterator const_group_list_iterator;


// calculate distance in meters between two lat/lon points
static double gc_dist( Point3D p1, Point3D p2 ) {
    Point3D r1( p1.x() * DEG_TO_RAD, p1.y() * DEG_TO_RAD, 0 );
    Point3D r2( p2.x() * DEG_TO_RAD, p2.y() * DEG_TO_RAD, 0 );

    // d=2*asin(sqrt((sin((lat1-lat2)/2))^2 + 
    //          cos(lat1)*cos(lat2)*(sin((lon1-lon2)/2))^2))
    double tmp1 = sin( (r1.y() - r2.y()) / 2.0 );
    double tmp2 = sin( (r1.x() - r2.x()) / 2.0 );

    // d=2*asin(sqrt((tmp1)^2 + cos(lat1)*cos(lat2)*(tmp2)^2))
    double clat1 = cos( r1.y() );
    double clat2 = cos( r1.y() );

    // d=2*asin(sqrt(tmp1*tmp1 + clat1*clat2*tmp2*tmp2))
    double tmp3 = sqrt(tmp1*tmp1 + clat1*clat2*tmp2*tmp2);

    // d=2*asin(tmp3)
    double d_rad = 2 * asin( tmp3 );
    // cout << "  dist (rad) = " << d_rad << endl;

    double d_nm = d_rad * RAD_TO_NM;
    // cout << "  dist (nm) = " << d_nm << endl;

    double d_m = d_nm * NM_TO_METER;
    // cout << "  dist (m) = " << d_m << endl;

    return d_m;
}


// calculate true course between two points given the distance in meters
static double gc_course( Point3D p1, Point3D p2, double d_m ) {
    double lon1 = p1.x() * DEG_TO_RAD;
    double lon2 = p2.x() * DEG_TO_RAD;
    double lat1 = p1.y() * DEG_TO_RAD;
    double lat2 = p2.y() * DEG_TO_RAD;

    double d_rad = d_m * METER_TO_NM * NM_TO_RAD;

    double tc1;

    if ( cos(lat1) < FG_EPSILON) {
	if ( lat1 > 0.0 ) {
	    tc1 = FG_PI;        //  starting from N pole
	} else {
	    tc1 = 0.0;		//  starting from S pole
	}
    }
	
    // For starting points other than the poles: 
	
    if ( sin(lon2 - lon1) < 0.0 ) {
	tc1 = acos( (sin(lat2)-sin(lat1)*cos(d_rad))/(sin(d_rad)*cos(lat1)));
    } else {
	tc1 = FG_2PI - 
	    acos((sin(lat2)-sin(lat1)*cos(d_rad))/(sin(d_rad)*cos(lat1)));
    }

    return tc1;
}


// calculate texture coordinates for a 1/2 runway.  Returns a mirror
// polygon to the runway, except each point is the texture coordinate
// of the corresponding point in the original polygon.
static FGPolygon rwy_calc_tex_coords( double lon, double lat, 
				      double length, double width, double hdg, 
				      FGPolygon in_poly )
{
    FGPolygon result;
    result.erase();

    Point3D center( lon, lat, 0 );
    double angle = hdg * DEG_TO_RAD;
    Point3D p, tp;
    double dist, course;
    double x, y, tx, ty;

    for ( int i = 0; i < in_poly.contours(); ++i ) {
	for ( int j = 0; j < in_poly.contour_size( i ); ++j ) {
	    p = in_poly.get_pt( i, j );
	    dist = gc_dist( center, p );
	    course = gc_course( center, p, dist ) + angle;
	    while ( course < -360.0 ) { course += 360.0; }
	    while ( course > 360.0 ) { course -= 360.0; }
	    // cout << "Dist = " << dist << endl;
	    // cout << "Course = " << course * 180.0 / FG_PI << endl;

	    x = cos( course ) * dist;
	    y = sin( course ) * dist;

	    tx = x / (length / 2.0);
	    tx = ((int)(tx * 100)) / 100.0;
	    if ( tx < -1.0 ) { tx = -1.0; }
	    if ( tx > 1.0 ) { tx = 1.0; }

	    ty = (y + (width / 2.0)) / width;
	    ty = ((int)(ty * 100)) / 100.0;
	    if ( ty < -1.0 ) { ty = -1.0; }
	    if ( ty > 1.0 ) { ty = 1.0; }

	    tp = Point3D( tx, ty, 0 );
	    // cout << "  (" << tx << "," << ty << ")" << endl;

	    result.add_node( i, tp );
	}
    }

    return result;
}


#if 0
// Find a the specified point in the polygon and set the contour/index
// values for it.  Returns true if a match found, false otherwise
static bool find_in_polygon( const Point3D p, const FGPolygon poly, 
			     int *contour, int *index ) {
    *contour = *index = -1;
    Point3D tmp;

    for ( int i = 0; i < poly.contours(); ++i ) {
	for ( int j = 0; j < poly.contour_size( i ); ++j ) {
	    tmp = poly.get_pt( i, j );
	    if ( tmp == p ) {
		*contour = i;
		*index = j;
		return true;
	    }
	}
    }

    return false;
}
#endif


#if 0
// Add points to keep line segment lengths under 1000'
static FGPolygon add_points( const FGPolygon& in_poly ) {
    FGPolygon result;
    result.erase();

    for ( int i = 0; i < in_poly.contours(); ++i ) {
	for ( int j = 0; j < in_poly.contour_size( i ) - 1; ++j ) {
	    gc_dist( in_poly.get_pt( i, j ),
		     in_poly.get_pt( i, j+1 ) );
	}
	gc_dist( in_poly.get_pt( i, in_poly.contour_size( i ) - 1 ),
		 in_poly.get_pt( i, 0 ) );
    }

    return result;
}
#endif


#if 0
// print polygon
static void print_poly( const FGPolygon& poly ) {
    for ( int i = 0; i < poly.contours(); ++i ) {
	cout << "contour " << i << endl;
 	for ( int j = 0; j < poly.contour_size( i ); ++j ) {
	    cout << "  " << poly.get_pt( i, j ) << endl;
	}
    }
}
#endif



// calculate the center of a list of points, by taking the halfway
// point between the min and max points.
Point3D calc_center( point_list& wgs84_nodes ) {
    Point3D p, min, max;

    if ( wgs84_nodes.size() ) {
	min = max = wgs84_nodes[0];
    } else {
	min = max = Point3D( 0 );
    }

    for ( int i = 0; i < (int)wgs84_nodes.size(); ++i ) {
	p = wgs84_nodes[i];

	if ( p.x() < min.x() ) { min.setx( p.x() ); }
	if ( p.y() < min.y() ) { min.sety( p.y() ); }
	if ( p.z() < min.z() ) { min.setz( p.z() ); }

	if ( p.x() > max.x() ) { max.setx( p.x() ); }
	if ( p.y() > max.y() ) { max.sety( p.y() ); }
	if ( p.z() > max.z() ) { max.setz( p.z() ); }
    }

    return ( min + max ) / 2.0;
}

// calculate the global bounding sphere.  Center is the center of the
// tile and zero elevation
double calc_bounding_radius( Point3D center, point_list& wgs84_nodes ) {
    double dist_squared;
    double radius_squared = 0;
    
    for ( int i = 0; i < (int)wgs84_nodes.size(); ++i ) {
        dist_squared = center.distance3Dsquared( wgs84_nodes[i] );
	if ( dist_squared > radius_squared ) {
            radius_squared = dist_squared;
        }
    }

    return sqrt(radius_squared);
}


// fix node elevations
point_list calc_elevations( const string& root, const point_list& geod_nodes ) {
    bool done = false;
    point_list result = geod_nodes;
    int i;
    FGArray array;

    // set all elevations to -9999
    for ( i = 0; i < (int)result.size(); ++i ) {
	result[i].setz( -9999.0 );
    }

    while ( !done ) {
	// find first node with -9999 elevation
	i = 0;
	while ( result[i].z() > -9000 ) { ++i; }

	if ( i < (int)result.size() ) {
	    FGBucket b( result[i].x(), result[i].y() );
	    string base = b.gen_base_path();

	    // try 3 arcsec dems first
	    string dem_path = root + "/DEM-3/Scenery/" + base 
		+ "/" + b.gen_index_str() + ".dem";
	    cout << "dem_path = " << dem_path << endl;
	
	    if ( ! array.open(dem_path) ) {
		cout << "ERROR: cannot open 3 arcsec file " << dem_path << endl;
		cout << "trying 30 arcsec file" << endl;
		
		// try 30 arcsec dem
		dem_path = root + "/DEM-30/Scenery/" + base 
		    + "/" + b.gen_index_str() + ".dem";
		cout << "dem_path = " << dem_path << endl;
		if ( ! array.open(dem_path) ) {
		    cout << "ERROR: cannot open 3 arcsec file " 
			 << dem_path << endl;
		}
	    }
	    array.parse( b );

	    // update all the non-updated elevations that are inside
	    // this dem file
	    double elev;
	    done = true;
	    for ( int j = 0; j < (int)result.size(); ++j ) {
		if ( result[j].z() < -9000 ) {
		    done = false;
		    cout << "interpolating for " << result[j] << endl;
		    elev = array.interpolate_altitude( result[j].x() * 3600.0,
						   result[j].y() * 3600.0 );
		    if ( elev > -9000 ) {
			result[j].setz( elev );
		    }
		}
	    }
	    array.close();
	} else {
	    done = true;
	}
    }


    return result;
}


// write out the structures to a file.  We assume that the groups come
// to us sorted by material property.  If not, things don't break, but
// the result may not be as efficient.
void write( const FGBucket& b, Point3D gbs_center, double gbs_radius,
	    const point_list& wgs84_nodes, const point_list& normals,
	    const point_list& texcoords, 
	    const group_list& strips_v, const group_list& strips_tc, 
	    const string_list& strip_materials,
	    const group_list& fans_v, const group_list& fans_tc,
	    const string_list& fan_materials )
{
    Point3D p;

    cout << "strip size = " << strips_v.size() << "  strip_materials = " 
	 << strip_materials.size() << endl;
    cout << "points = " << wgs84_nodes.size() << endl;
    cout << "tex coords = " << texcoords.size() << endl;
    // write headers
    printf("# FGFS Scenery\n");
    printf("# Version %s\n", FG_SCENERY_FILE_FORMAT);

    time_t calendar_time = time(NULL);
    struct tm *local_tm;
    local_tm = localtime( &calendar_time );
    char time_str[256];
    strftime( time_str, 256, "%a %b %d %H:%M:%S %Z %Y", local_tm);
    printf("# Created %s\n", time_str );
    printf("\n");

    // write global bounding sphere
    printf("# gbs %.5f %.5f %.5f %.2f\n",
	   gbs_center.x(), gbs_center.y(), gbs_center.z(), gbs_radius);
    printf("\n");

    // dump vertex list
    printf("# vertex list\n");
    for ( int i = 0; i < (int)wgs84_nodes.size(); ++i ) {
	p = wgs84_nodes[i] - gbs_center;
	
	printf( "v %.5f %.5f %.5f\n", p.x(), p.y(), p.z() );
    }
    printf("\n");

    printf("# vertex normal list\n");
    for ( int i = 0; i < (int)normals.size(); ++i ) {
	p = normals[i];
	printf( "vn %.5f %.5f %.5f\n", p.x(), p.y(), p.z() );
    }
    printf("\n");

    // dump texture coordinates
    printf("# texture coordinate list\n");
    for ( int i = 0; i < (int)texcoords.size(); ++i ) {
	p = texcoords[i];
	printf( "vt %.5f %.5f\n", p.x(), p.y() );
    }
    printf("\n");

    // dump triangle groups
    printf("# triangle groups\n");

    int start = 0;
    int end = 1;
    string material;
    while ( start < (int)strip_materials.size() ) {
	// find next group
	material = strip_materials[start];
	while ( (end < (int)strip_materials.size()) && 
		(material == strip_materials[end]) )
	{
	    // cout << "end = " << end << endl;
	    end++;
	}
	// cout << "group = " << start << " to " << end - 1 << endl;

	// make a list of points for the group
	point_list group_nodes;
	group_nodes.clear();
	Point3D bs_center;
	double bs_radius;
	for ( int i = start; i < end; ++i ) {
	    for ( int j = 0; j < (int)strips_v[i].size(); ++j ) {
		group_nodes.push_back( wgs84_nodes[ strips_v[i][j] ] );
		bs_center = calc_center( group_nodes );
		bs_radius = calc_bounding_radius( bs_center, group_nodes );
	    }
	}

	// write group headers
	printf("\n");
	printf("# usemtl %s\n", material.c_str());
	printf("# bs %.4f %.4f %.4f %.2f\n",
	       bs_center.x(), bs_center.y(), bs_center.z(), bs_radius);

	// write groups
	for ( int i = start; i < end; ++i ) {
	    printf("ts");
	    for ( int j = 0; j < (int)strips_v[i].size(); ++j ) {
		printf(" %d/%d", strips_v[i][j], strips_tc[i][j] );
	    }
	    printf("\n");
	}

	start = end;
	end = start + 1;
    }
}

    
// build 3d airport
void build_airport( string airport_raw, list < string > & runway_list,
		    const string& root ) {

    poly_list rwy_nodes, rwy_strips, rwy_txs;
    FGPolygon runway, runway_a, runway_b, result, result_a, result_b;
    FGPolygon strip_a, strip_b, tc_a, tc_b;
    FGPolygon base;
    point_list apt_pts;
    Point3D p;

    FGPolygon accum;
    accum.erase();

    // parse main airport information
    double apt_lon, apt_lat;
    int elev;

    cout << airport_raw << endl;
    string apt_type = airport_raw.substr(0, 1);
    string apt_code = airport_raw.substr(2, 4);
    string apt_lat_str = airport_raw.substr(7, 10);
    apt_lat = atof( apt_lat_str.c_str() );
    string apt_lon_str = airport_raw.substr(18, 11);
    apt_lon = atof( apt_lon_str.c_str() );
    string apt_elev = airport_raw.substr(30, 5);
    elev = atoi( apt_elev.c_str() );
    string apt_use = airport_raw.substr(36, 1);
    string apt_twr = airport_raw.substr(37, 1);
    string apt_bldg = airport_raw.substr(38, 1);
    string apt_name = airport_raw.substr(40);

    /*
    cout << "  type = " << apt_type << endl;
    cout << "  code = " << apt_code << endl;
    cout << "  lat  = " << apt_lat << endl;
    cout << "  lon  = " << apt_lon << endl;
    cout << "  elev = " << apt_elev << " " << elev << endl;
    cout << "  use  = " << apt_use << endl;
    cout << "  twr  = " << apt_twr << endl;
    cout << "  bldg = " << apt_bldg << endl;
    cout << "  name = " << apt_name << endl;
    */

    FGBucket b( apt_lon, apt_lat );
    Point3D center_geod( b.get_center_lon() * DEG_TO_RAD,
			 b.get_center_lat() * DEG_TO_RAD, 0 );
    Point3D gbs_center = fgGeodToCart( center_geod );
    cout << b.gen_base_path() << "/" << b.gen_index_str() << endl;

    // Ignore any seaplane bases
    if ( apt_type == "S" ) {
	return;
    }

    // parse runways and generate the vertex list
    string rwy_str;
    double lon, lat, hdg;
    int len, width;

    list < string >::iterator last_runway = runway_list.end();
    list < string >::iterator current_runway = runway_list.begin();
    for ( ; current_runway != last_runway ; ++current_runway ) {
	rwy_str = (*current_runway);

	cout << rwy_str << endl;
	string rwy_no = rwy_str.substr(2, 4);
	string rwy_lat = rwy_str.substr(6, 10);
	lat = atof( rwy_lat.c_str() );
	string rwy_lon = rwy_str.substr(17, 11);
	lon = atof( rwy_lon.c_str() );
	string rwy_hdg = rwy_str.substr(29, 7);
	hdg = atof( rwy_hdg.c_str() );
	string rwy_len = rwy_str.substr(36, 7);
	len = atoi( rwy_len.c_str() );
	string rwy_width = rwy_str.substr(43, 4);
	width = atoi( rwy_width.c_str() );
	string rwy_sfc = rwy_str.substr(47, 4);
	string rwy_end1 = rwy_str.substr(52, 8);
	string rwy_end2 = rwy_str.substr(61, 8);

	/*
	cout << "  no    = " << rwy_no << endl;
	cout << "  lat   = " << rwy_lat << " " << lat << endl;
	cout << "  lon   = " << rwy_lon << " " << lon << endl;
	cout << "  hdg   = " << rwy_hdg << " " << hdg << endl;
	cout << "  len   = " << rwy_len << " " << len << endl;
	cout << "  width = " << rwy_width << " " << width << endl;
	cout << "  sfc   = " << rwy_sfc << endl;
	cout << "  end1  = " << rwy_end1 << endl;
	cout << "  end2  = " << rwy_end2 << endl;
	*/

	runway = gen_runway_w_mid( lon, lat, hdg * DEG_TO_RAD, 
				   (double)len * FEET_TO_METER,
				   (double)width * FEET_TO_METER );

	// runway half "a"
	runway_a.erase();
	runway_a.add_node( 0, runway.get_pt(0, 0) );
	runway_a.add_node( 0, runway.get_pt(0, 1) );
	runway_a.add_node( 0, runway.get_pt(0, 2) );
	runway_a.add_node( 0, runway.get_pt(0, 5) );

	// runway half "b"
	runway_b.erase();
	runway_b.add_node( 0, runway.get_pt(0, 5) );
	runway_b.add_node( 0, runway.get_pt(0, 2) );
	runway_b.add_node( 0, runway.get_pt(0, 3) );
	runway_b.add_node( 0, runway.get_pt(0, 4) );
	
	cout << "raw runway pts (a half)" << endl;
	for ( int j = 0; j < runway_a.contour_size( 0 ); ++j ) {
	    p = runway_a.get_pt(0, j);
	    cout << " point = " << p << endl;
	}
	cout << "raw runway pts (b half)" << endl;
	for ( int j = 0; j < runway_b.contour_size( 0 ); ++j ) {
	    p = runway_b.get_pt(0, j);
	    cout << " point = " << p << endl;
	}
	
	result_a = polygon_diff( runway_a, accum );
	strip_a = polygon_to_tristrip( result_a );

	cout << "result_a = " << result_a.contours() << endl;
	accum = polygon_union( runway_a, accum );
	tc_a = rwy_calc_tex_coords( lon, lat, 
				    (double)len * FEET_TO_METER,
				    (double)width * FEET_TO_METER, 
				    hdg, strip_a );
	rwy_nodes.push_back( result_a );
	rwy_strips.push_back( strip_a );
	rwy_txs.push_back( tc_a );

	result_b = polygon_diff( runway_b, accum );
	strip_b = polygon_to_tristrip( result_b );

	cout << "result_b = " << result_b.contours() << endl;
	accum = polygon_union( runway_b, accum );
	tc_b = rwy_calc_tex_coords( lon, lat,
				    (double)len * FEET_TO_METER,
				    (double)width * FEET_TO_METER, 
				    hdg + 180.0, strip_b );
	rwy_nodes.push_back( result_b );
	rwy_strips.push_back( strip_b );
	rwy_txs.push_back( tc_b );

	// print runway points
	cout << "clipped runway pts (a)" << endl;
	for ( int i = 0; i < result_a.contours(); ++i ) {
	    for ( int j = 0; j < result_a.contour_size( i ); ++j ) {
		p = result_a.get_pt(i, j);
		cout << " point = " << p << endl;
	    }
	}

	// print runway points
	cout << "clipped runway pts (b)" << endl;
	for ( int i = 0; i < result_b.contours(); ++i ) {
	    for ( int j = 0; j < result_b.contour_size( i ); ++j ) {
		p = result_b.get_pt(i, j);
		cout << " point = " << p << endl;
	    }
	}

	base = gen_runway_area( lon, lat, hdg * DEG_TO_RAD, 
				(double)(len+40) * FEET_TO_METER,
				(double)(width+40) * FEET_TO_METER );

	// add base to apt_pts (for convex hull of airport area)
	for ( int i = 0; i < base.contour_size( 0 ); ++i ) {
	    p = base.get_pt(0, i);
	    // cout << " point = " << p << endl;
	    apt_pts.push_back( p );
	}
    }

    if ( apt_pts.size() == 0 ) {
	cout << "no airport points generated" << endl;
	return;
    }

    // generate convex hull and strip version
    FGPolygon hull = convex_hull(apt_pts);
    FGPolygon base_nodes = polygon_diff( hull, accum );
    FGPolygon base_strips = polygon_to_tristrip( base_nodes );

    //
    // We should now have the runway polygons all generated with their
    // corresponding strips and texture coordinates, and the
    // surrounding base area.
    //
    // Next we need to create the output lists ... vertices, normals,
    // texture coordinates, and tri-strips
    //

    // traverse the strip list and create ordered node and texture
    // coordinate lists

    FGTriNodes nodes, texcoords;
    nodes.clear();
    texcoords.clear();

    group_list strips_v; strips_v.clear();
    group_list strips_tc; strips_tc.clear();
    string_list strip_materials; strip_materials.clear();

    Point3D tc;
    int index;
    int_list strip_v;
    int_list strip_tc;

    for ( int k = 0; k < (int)rwy_strips.size(); ++k ) {
	cout << "strip " << k << endl;
	FGPolygon strip_poly = rwy_strips[k];
	for ( int i = 0; i < strip_poly.contours(); ++i ) {
	    strip_v.clear();
	    strip_tc.clear();
	    for ( int j = 0; j < strip_poly.contour_size(i); ++j ) {
		p = strip_poly.get_pt( i, j );
		index = nodes.unique_add( p );
		strip_v.push_back( index );
		tc = rwy_txs[k].get_pt( i, j );
		index = texcoords.unique_add( tc );
		strip_tc.push_back( index );
	    }
	    strips_v.push_back( strip_v );
	    strips_tc.push_back( strip_tc );
	    strip_materials.push_back( "Concrete" );
	}
    }
    
    // add base points
    point_list base_txs; 
    int_list base_tc;
    for ( int i = 0; i < base_strips.contours(); ++i ) {
	strip_v.clear();
	strip_tc.clear();
	for ( int j = 0; j < base_strips.contour_size(i); ++j ) {
	    p = base_strips.get_pt( i, j );
	    index = nodes.unique_add( p );
	    strip_v.push_back( index );
	}
	strips_v.push_back( strip_v );
	strip_materials.push_back( "Grass" );

	base_txs.clear();
	base_txs = calc_tex_coords( b, nodes.get_node_list(), strip_v );

	base_tc.clear();
	for ( int j = 0; j < (int)base_txs.size(); ++j ) {
	    tc = base_txs[j];
	    cout << "base_tc = " << tc << endl;
	    index = texcoords.simple_add( tc );
	    base_tc.push_back( index );
	}
	strips_tc.push_back( base_tc );
    }

    // calculate node elevations
    point_list geod_nodes = calc_elevations( root, nodes.get_node_list() );

    // calculate wgs84 mapping of nodes
    point_list wgs84_nodes;
    for ( int i = 0; i < (int)geod_nodes.size(); ++i ) {
	p.setx( geod_nodes[i].x() * DEG_TO_RAD );
	p.sety( geod_nodes[i].y() * DEG_TO_RAD );
	p.setz( geod_nodes[i].z() );
	wgs84_nodes.push_back( fgGeodToCart( p ) );
    }
    double gbs_radius = calc_bounding_radius( gbs_center, wgs84_nodes );

    // calculate normal(s) for this airport
    p.setx( rwy_strips[0].get_pt(0, 0).x() * DEG_TO_RAD );
    p.sety( rwy_strips[0].get_pt(0, 0).y() * DEG_TO_RAD );
    p.setz( 0 );
    Point3D tmp = fgGeodToCart( p );
    // cout << "geod = " << p << endl;
    // cout << "cart = " << tmp << endl;

    sgdVec3 vn;
    sgdSetVec3( vn, tmp.x(), tmp.y(), tmp.z() );
    sgdNormalizeVec3( vn );
    point_list normals;
    normals.clear();
    for ( int i = 0; i < (int)nodes.size(); ++i ) {
	normals.push_back( Point3D( vn[0], vn[1], vn[2] ) );
    }

    group_list fans_v; fans_v.clear();
    group_list fans_tc; fans_tc.clear();
    string_list fan_materials;
    fan_materials.clear();

    write( b, gbs_center, gbs_radius, wgs84_nodes, normals,
	   texcoords.get_node_list(), 
	   strips_v, strips_tc, strip_materials, 
	   fans_v, fans_tc, fan_materials );
}
